:root {
    --hcipy-primary: #455265;/*#27ae60;*/
    --hcipy-secondary: #2c3645;
    --hcipy-tertiary: #1e2530;
	--hcipy-accent: #2980b9;
}

div.tutorial-item {
	border: none;
	border-left: 10px solid var(--hcipy-accent);
	border-radius: 10px;
	box-shadow: 3px 2px 8px 2px rgba(0, 0, 0, 0.3);
	margin-bottom: 20px;
	padding: 10px;
	display: block;
}

div.tutorial-item:hover {
	box-shadow: 1px 1px 2px 1px rgba(0, 0, 0, 0.2);
}

div.tutorial-item a {
	font-size: 125%;
	font-weight: 700;
	color: black;
	padding-left: 10px;
	padding-top: 10px;
}

div.tutorial-item p {
	margin-bottom: 15px;
}

div.tutorial-row {
	display: flex;
	flex-wrap: wrap;
}

div.tutorial-thumbnail {
	padding: 0px 3px 0px 0px;
	display: block;
	max-width: 25%;
	min-width: 100px;
}

div.tutorial-thumbnail img {
	vertical-align: middle;
}

div.tutorial-description {
	display: block;
	max-width: 75%;
	padding-left: 20px;
}

.wy-side-nav-search, .wy-nav-top {
	background-color: var(--hcipy-secondary);
}

.wy-nav-side {
	background-color: var(--hcipy-secondary);
}

.wy-menu-vertical a:hover {
	background-color: var(--hcipy-primary);
}

.wy-menu-vertical p.caption {
	color: var(--hcipy-accent);
}

.rst-content a:link, .rst-content a:visited {
	color: var(--hcipy-accent);
}

body {
	color: var(--hcipy-tertiary);
}

.wy-nav-content {
	max-width: 900px;
	margin: auto;
}

.rst-content video {
	max-width: 100%;
}

.rst-content blockquote {
	border-left: 4px solid var(--hcipy-accent);
	padding-left: 16px;
	margin-left: 0;
	font-style: italic;
}
